@@ -13,6 +13,7 @@ use webcore::number::Number;
 use webcore::type_name::type_name;
 use webcore::symbol::Symbol;
 use webcore::unsafe_typed_array::UnsafeTypedArray;
+use webcore::mutfn::Mut;
 use webcore::once::Once;
 use webcore::global_arena;
 
@@ -20,9 +21,12 @@ use webcore::value::{
     Null,
     Undefined,
     Reference,
-    Value
+    Value,
+    ConversionError
 };
 
+use webapi::error::TypeError;
+
 #[repr(u8)]
 #[derive(Copy, Clone, PartialEq, Debug)]
 pub enum Tag {
@@ -37,6 +41,7 @@ pub enum Tag {
     Object = 8,
     Reference = 9,
     Function = 10,
+    FunctionMut = 12,
     FunctionOnce = 13,
     UnsafeTypedArray = 14,
     Symbol = 15
@@ -149,6 +154,14 @@ struct SerializedUntaggedFunction {
     deallocator_pointer: u32
 }
 
+#[repr(C)]
+#[derive(Debug)]
+struct SerializedUntaggedFunctionMut {
+    adapter_pointer: u32,
+    pointer: u32,
+    deallocator_pointer: u32
+}
+
 #[repr(C)]
 #[derive(Debug)]
 struct SerializedUntaggedFunctionOnce {
@@ -386,6 +399,7 @@ untagged_boilerplate!( test_array, as_array, Tag::Array, SerializedUntaggedArray
 untagged_boilerplate!( test_symbol, as_symbol, Tag::Symbol, SerializedUntaggedSymbol );
 untagged_boilerplate!( test_reference, as_reference, Tag::Reference, SerializedUntaggedReference );
 untagged_boilerplate!( test_function, as_function, Tag::Function, SerializedUntaggedFunction );
+untagged_boilerplate!( test_function_mut, as_function_mut, Tag::FunctionMut, SerializedUntaggedFunctionMut );
 untagged_boilerplate!( test_function_once, as_function_once, Tag::FunctionOnce, SerializedUntaggedFunctionOnce );
 untagged_boilerplate!( test_unsafe_typed_array, as_unsafe_typed_array, Tag::UnsafeTypedArray, SerializedUntaggedUnsafeTypedArray );
 
@@ -404,6 +418,7 @@ impl< 'a > SerializedValue< 'a > {
             Tag::Reference => self.as_reference().deserialize().into(),
             Tag::Symbol => self.as_symbol().deserialize().into(),
             Tag::Function |
+            Tag::FunctionMut |
             Tag::FunctionOnce |
             Tag::Object |
             Tag::Array |
@@ -756,18 +771,24 @@ trait FuncallAdapter< F > {
     extern fn deallocator( callback: *mut F );
 }
 
-macro_rules! impl_for_fn {
-    ($next:tt => $($kind:ident),*) => {
-        impl< $($kind: TryFrom< Value >,)* F > FuncallAdapter< F > for Newtype< (FunctionTag, ($($kind,)*)), F >
-            where F: CallMut< ($($kind,)*) > + 'static, F::Output: JsSerializeOwned
+macro_rules! impl_for_fn_and_modifier {
+    (
+        args: ($($kind:ident),*),
+        trait: $trait:ident,
+        wrapped type: $wrappedtype:ty,
+        unwrap: $wrapped:ident => $unwrap:expr,
+        serialized to: $serialized_to:tt,
+        call: $callback:ident => $call:expr
+    ) => {
+        impl< $($kind: TryFrom< Value >,)* F > FuncallAdapter< F > for Newtype< (FunctionTag, ($($kind,)*)), $wrappedtype >
+            where F: $trait< ($($kind,)*) > + 'static, F::Output: JsSerializeOwned
         {
             #[allow(unused_mut, unused_variables, non_snake_case)]
             extern fn funcall_adapter(
-                    callback: *mut F,
+                    $callback: *mut F,
                     raw_arguments: *mut SerializedUntaggedArray
                 )
             {
-                let callback = unsafe { &mut *callback };
                 let mut arguments = unsafe { &*raw_arguments }.deserialize();
 
                 unsafe {
@@ -798,7 +819,7 @@ macro_rules! impl_for_fn {
 
                 $crate::private::noop( &mut nth_argument );
 
-                let result = callback.call_mut( ($($kind,)*) );
+                let result = $call;
 
                 let mut result = Some( result );
                 let result = JsSerializeOwned::into_js_owned( &mut result );
@@ -817,78 +838,16 @@ macro_rules! impl_for_fn {
             }
         }
 
-        impl< $($kind: TryFrom< Value >,)* F > FuncallAdapter< F > for Newtype< (FunctionTag, ($($kind,)*)), Once< F > >
-            where F: CallOnce< ($($kind,)*) > + 'static, F::Output: JsSerializeOwned
-        {
-            #[allow(unused_mut, unused_variables, non_snake_case)]
-            extern fn funcall_adapter(
-                    callback: *mut F,
-                    raw_arguments: *mut SerializedUntaggedArray
-                )
-            {
-                debug_assert_ne!( callback, 0 as *mut F );
-
-                let callback = unsafe {
-                    Box::from_raw( callback )
-                };
-
-                let mut arguments = unsafe { &*raw_arguments }.deserialize();
-                unsafe {
-                    ffi::dealloc( raw_arguments as *mut u8, mem::size_of::< SerializedValue >() );
-                }
-
-                if arguments.len() != F::expected_argument_count() {
-                    // TODO: Should probably throw an exception into the JS world or something like that.
-                    panic!( "Expected {} arguments, got {}", F::expected_argument_count(), arguments.len() );
-                }
-
-                let mut arguments = arguments.drain( .. );
-                let mut nth_argument = 0;
-                $(
-                    let $kind = match arguments.next().unwrap().try_into() {
-                        Ok( value ) => value,
-                        Err( _ ) => {
-                            panic!(
-                                "Argument #{} is not convertible to '{}'",
-                                nth_argument + 1,
-                                type_name::< $kind >()
-                            );
-                        }
-                    };
-
-                    nth_argument += 1;
-                )*
-
-                $crate::private::noop( &mut nth_argument );
-
-                let result = callback.call_once( ($($kind,)*) );
-
-                let mut result = Some( result );
-                let result = JsSerializeOwned::into_js_owned( &mut result );
-                let result = &result as *const _;
-
-                // This is kinda hacky but I'm not sure how else to do it at the moment.
-                __js_raw_asm!( "Module.STDWEB_PRIVATE.tmp = Module.STDWEB_PRIVATE.to_js( $0 );", result );
-            }
-
-            extern fn deallocator( callback: *mut F ) {
-                let callback = unsafe {
-                    Box::from_raw( callback )
-                };
-
-                drop( callback );
-            }
-        }
-
-        impl< $($kind: TryFrom< Value >,)* F > JsSerializeOwned for Newtype< (FunctionTag, ($($kind,)*)), Once< F > >
-            where F: CallOnce< ($($kind,)*) > + 'static, F::Output: JsSerializeOwned
+        impl< $($kind: TryFrom< Value >,)* F > JsSerializeOwned for Newtype< (FunctionTag, ($($kind,)*)), $wrappedtype >
+            where F: $trait< ($($kind,)*) > + 'static, F::Output: JsSerializeOwned
         {
             #[inline]
             fn into_js_owned< 'a >( value: &'a mut Option< Self > ) -> SerializedValue< 'a > {
-                let callback: *mut F = Box::into_raw( Box::new( value.take().unwrap().unwrap_newtype().0 ) );
+                let $wrapped = value.take().unwrap().unwrap_newtype();
+                let callback: *mut F = Box::into_raw( Box::new( $unwrap ) );
                 let adapter_pointer = <Self as FuncallAdapter< F > >::funcall_adapter;
                 let deallocator_pointer = <Self as FuncallAdapter< F > >::deallocator;
-                SerializedUntaggedFunctionOnce {
+                $serialized_to {
                     adapter_pointer: adapter_pointer as u32,
                     pointer: callback as u32,
                     deallocator_pointer: deallocator_pointer as u32
@@ -896,32 +855,16 @@ macro_rules! impl_for_fn {
             }
         }
 
-        impl< $($kind: TryFrom< Value >,)* F > JsSerializeOwned for Newtype< (FunctionTag, ($($kind,)*)), F >
-            where F: CallMut< ($($kind,)*) > + 'static, F::Output: JsSerializeOwned
+        impl< $($kind: TryFrom< Value >,)* F > JsSerializeOwned for Newtype< (FunctionTag, ($($kind,)*)), Option< $wrappedtype > >
+            where F: $trait< ($($kind,)*) > + 'static, F::Output: JsSerializeOwned
         {
             #[inline]
             fn into_js_owned< 'a >( value: &'a mut Option< Self > ) -> SerializedValue< 'a > {
-                let callback: *mut F = Box::into_raw( Box::new( value.take().unwrap().unwrap_newtype() ) );
-                let adapter_pointer = <Self as FuncallAdapter< F > >::funcall_adapter;
-                let deallocator_pointer = <Self as FuncallAdapter< F > >::deallocator;
-                SerializedUntaggedFunction {
-                    adapter_pointer: adapter_pointer as u32,
-                    pointer: callback as u32,
-                    deallocator_pointer: deallocator_pointer as u32
-                }.into()
-            }
-        }
-
-        impl< $($kind: TryFrom< Value >,)* F > JsSerializeOwned for Newtype< (FunctionTag, ($($kind,)*)), Option< F > >
-            where F: CallMut< ($($kind,)*) > + 'static, F::Output: JsSerializeOwned
-        {
-            #[inline]
-            fn into_js_owned< 'a >( value: &'a mut Option< Self > ) -> SerializedValue< 'a > {
-                if let Some( value ) = value.take().unwrap().unwrap_newtype() {
-                    let callback: *mut F = Box::into_raw( Box::new( value ) );
-                    let adapter_pointer = <Newtype< (FunctionTag, ($($kind,)*)), F > as FuncallAdapter< F > >::funcall_adapter;
-                    let deallocator_pointer = <Newtype< (FunctionTag, ($($kind,)*)), F > as FuncallAdapter< F > >::deallocator;
-                    SerializedUntaggedFunction {
+                if let Some( $wrapped ) = value.take().unwrap().unwrap_newtype() {
+                    let callback: *mut F = Box::into_raw( Box::new( $unwrap ) );
+                    let adapter_pointer = <Newtype< (FunctionTag, ($($kind,)*)), $wrappedtype > as FuncallAdapter< F > >::funcall_adapter;
+                    let deallocator_pointer = <Newtype< (FunctionTag, ($($kind,)*)), $wrappedtype > as FuncallAdapter< F > >::deallocator;
+                    $serialized_to {
                         adapter_pointer: adapter_pointer as u32,
                         pointer: callback as u32,
                         deallocator_pointer: deallocator_pointer as u32
@@ -931,6 +874,37 @@ macro_rules! impl_for_fn {
                 }
             }
         }
+    }
+}
+
+macro_rules! impl_for_fn {
+    ($next:tt => $($kind:ident),*) => {
+        impl_for_fn_and_modifier!(
+            args: ($($kind),*),
+            trait: CallMut,
+            wrapped type: F,
+            unwrap: f => f,
+            serialized to: SerializedUntaggedFunction,
+            call: f => { unsafe { &mut *f }.call_mut( ($($kind,)*) ) }
+        );
+
+        impl_for_fn_and_modifier!(
+            args: ($($kind),*),
+            trait: CallMut,
+            wrapped type: Mut<F>,
+            unwrap: f => {f.0},
+            serialized to: SerializedUntaggedFunctionMut,
+            call: f => { unsafe { &mut *f }.call_mut( ($($kind,)*) ) }
+        );
+
+        impl_for_fn_and_modifier!(
+            args: ($($kind),*),
+            trait: CallOnce,
+            wrapped type: Once<F>,
+            unwrap: f => {f.0},
+            serialized to: SerializedUntaggedFunctionOnce,
+            call: f => { unsafe { Box::from_raw( f ) }.call_once( ($($kind,)*) ) }
+        );
 
         next! { $next }
     }
@@ -946,10 +920,20 @@ impl< 'a, T: ?Sized + JsSerialize > JsSerialize for &'a T {
     }
 }
 
+impl JsSerialize for ConversionError {
+    #[doc(hidden)]
+    fn _into_js< 'x >( &'x self ) -> SerializedValue< 'x > {
+        let type_error: TypeError = self.into();
+        let reference: Reference = type_error.into();
+        let value: Value = reference.into();
+        global_arena::serialize_value( value )
+    }
+}
+
 #[cfg(test)]
 mod test_deserialization {
     use std::rc::Rc;
-    use std::cell::Cell;
+    use std::cell::{Cell, RefCell};
     use super::*;
 
     #[test]
@@ -1156,6 +1140,20 @@ mod test_deserialization {
         assert_eq!( result, Value::String( "Dog!".to_owned() ) );
     }
 
+    #[test]
+    fn function_mut() {
+        let mut count = 0;
+        let callback = move || -> i32 {
+            count += 1;
+            count
+        };
+        let callback = js! { return @{Mut(callback)}; };
+        assert_eq!({ let x : i32 = js!{ return @{&callback}(); }.try_into().unwrap(); x }, 1);
+        assert_eq!({ let x : i32 = js!{ return @{&callback}(); }.try_into().unwrap(); x }, 2);
+        assert_eq!({ let x : i32 = js!{ return @{&callback}(); }.try_into().unwrap(); x }, 3);
+        js!{ @{callback}.drop(); };
+    }
+
     #[test]
     fn function_once_cannot_be_called_twice() {
         fn call< F: FnOnce() + 'static >( callback: F ) -> Value {
@@ -1233,6 +1231,179 @@ mod test_deserialization {
         let result = call( move || {} );
         assert_eq!( result, Value::Bool( true ) );
     }
+
+    #[test]
+    fn issue_273() {
+        let mut count = 0;
+        let f = move |callback: ::stdweb::Value| {
+            count += 1;
+            js! {
+                @{callback}();
+            };
+        };
+
+        let result = js! {
+            let f = @{Mut(f)};
+
+            let caught = false;
+
+            try {
+                f(function () {
+                    f(function() {});
+                });
+            } catch ( error ) {
+                if( error instanceof ReferenceError ) {
+                    caught = true;
+                }
+            }
+
+            f.drop();
+
+            return caught;
+        };
+        assert_eq!( result, Value::Bool( true ) );
+    }
+
+    #[test]
+    fn issue_277() {
+        struct MyStruct {
+            was_dropped: bool
+        }
+        impl MyStruct {
+            fn consume(self) {}
+        }
+        impl Drop for MyStruct {
+            fn drop(&mut self) {
+                assert_eq!(self.was_dropped, false);
+                self.was_dropped = true;
+            }
+        }
+
+        let s = MyStruct { was_dropped: false };
+
+        let f = move || -> () {
+            s.consume();
+            unreachable!(); // never actually called
+        };
+
+        js! {
+            let f = @{Once(f)};
+
+            let drop = f.drop;
+            drop();
+            drop();
+        };
+    }
+
+    #[test]
+    fn test_closure_dropped_while_being_called_is_dropped_after_it_returns() {
+        struct MarkTrueOnDrop( Rc< Cell< bool > > );
+        impl Drop for MarkTrueOnDrop {
+            fn drop( &mut self ) {
+                self.0.set( true );
+            }
+        }
+
+        let was_dropped = Rc::new( Cell::new( false ) );
+        let was_dropped_clone = was_dropped.clone();
+        let callback = move |itself: Value, check_if_dropped: Value| {
+            let _mark_true_on_drop = MarkTrueOnDrop( was_dropped_clone.clone() );
+            js!(
+                @{itself}.drop();
+                @{check_if_dropped}();
+            );
+        };
+
+        let check_if_dropped = move || {
+            assert_eq!( was_dropped.get(), false );
+        };
+
+        js!(
+            var callback = @{callback};
+            callback( callback, @{Once( check_if_dropped )} );
+        );
+    }
+
+    #[test]
+    fn test_dropping_the_closure_while_it_is_being_called_will_make_future_calls_throw() {
+        #[derive(Clone, Debug, PartialEq, Eq, ReferenceType)]
+        #[reference(instance_of = "ReferenceError")]
+        pub struct ReferenceError( Reference );
+
+        let counter = Rc::new( Cell::new( 0 ) );
+        let counter_clone = counter.clone();
+        let caught_value = Rc::new( RefCell::new( Value::Null ) );
+        let caught_value_clone = caught_value.clone();
+        let callback = move |itself: Value| {
+            let value = counter_clone.get();
+            counter_clone.set( value + 1 );
+
+            if value == 0 {
+                let caught = js!(
+                    var callback = @{itself};
+                    callback.drop();
+
+                    var caught = null;
+                    try {
+                        callback( callback );
+                    } catch( error ) {
+                        caught = error;
+                    }
+
+                    return caught;
+                );
+                *caught_value_clone.borrow_mut() = caught;
+            }
+        };
+
+        js!(
+            var callback = @{callback};
+            callback( callback );
+        );
+
+        assert_eq!( counter.get(), 1 );
+        let reference_error: Result< ReferenceError, _ > = caught_value.borrow().clone().try_into();
+        assert!( reference_error.is_ok() );
+    }
+
+    #[test]
+    fn no_return() {
+        let values = Rc::new( RefCell::new( Vec::new() ) );
+        let values_clone = values.clone();
+        let callback = move |value: i32| {
+            values_clone.borrow_mut().push( value );
+        };
+
+        let _: () = js!( @(no_return)
+            var cb = @{Mut( callback.clone() )};
+            cb( 1 );
+            cb.drop();
+        );
+
+        let _: () = js!( @(no_return)
+            var cb = @{Mut( callback.clone() )};
+            cb( 2 );
+            cb.drop();
+            return 20;
+        );
+
+        let a: Value = js!(
+            var cb = @{Mut( callback.clone() )};
+            cb( 3 );
+            cb.drop();
+        );
+
+        let b: Value = js!(
+            var cb = @{Mut( callback )};
+            cb( 4 );
+            cb.drop();
+            return 40;
+        );
+
+        assert_eq!( *values.borrow(), &[1, 2, 3, 4] );
+        assert_eq!( a, Value::Undefined );
+        assert_eq!( b, 40 )
+    }
 }
 
 #[cfg(test)]
